#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <err.h>

struct tree_node
{   u_int32_t mask;
    u_int32_t value;
    u_int32_t eq;
    u_int32_t neq;
};

int flag_v;
struct tree_node *ip_tree;

void print_ip(FILE *f, u_int32_t ip)
{   fprintf(f, "%3u", (ip >> 24) & 0xFF);
    fprintf(f, ".%u", (ip >> 16) & 0xFF);
    fprintf(f, ".%u", (ip >> 8) & 0xFF);
    fprintf(f, ".%u", ip & 0xFF);
}

void usage(int status)
{   fprintf( stderr
           , "traf-policy @BABOLO V.M "VERS"  "DATE"\n"
             "   Usage:\n"
             "traf-policy -t|-d [-i InFile] [-o Outfile]\n"
           );
    exit(status);
}

int main(int argc, char **argv)
{   int c, ifile, ofile, fun;
    size_t n;
    FILE *f = NULL;

    ifile = fileno(stdin);
    ofile = fileno(stdout);
    fun = 0;
    while ((c = getopt(argc, argv, "vtdi:o:h")) > 0)
    {   switch (c)
        {case 'v':
            flag_v++;
            break;
         case 't':
            fun++;
            break;
         case 'd':
            fun--;
            break;
         case 'i':
            if  ((ifile = open(optarg, O_RDONLY)) < 0)        err(EX_NOINPUT, "Can't open %s", optarg);
            break;
         case 'o':
            if  ((ofile = open(optarg, O_WRONLY | O_CREAT | O_TRUNC | O_EXLOCK, 0666)) < 0)
                                                              err(EX_NOINPUT, "Can't open %s", optarg);
            break;
         case 'h': usage(EX_OK);
         default: usage(EX_USAGE);
    }   }

    argc -= optind;
    argv += optind;

    if  (fun == 0)
    {   fprintf(stderr, "What to do?\n");
        usage(EX_USAGE);
    }else if (fun < 0)
    {   struct stat sb;
        size_t len_tree;
        if  (fstat(ifile, &sb) < 0)                           err(EX_NOINPUT, "Can't stat input file");
        ip_tree = (struct tree_node*)mmap( NULL
                                         , sb.st_size
                                         , PROT_READ
                                         , 0
                                         , ifile
                                         , 0
                                         );
        if  ((void*)ip_tree == MAP_FAILED)                    err(EX_NOINPUT, "Can't mmap input file");
        len_tree = sb.st_size / sizeof(struct tree_node);
        close(ifile);
        if  ((f = fdopen(ofile, "wb")) < 0)                err(EX_NOINPUT, "Can't fdopen output file");
        for (n = 0; n < len_tree; n++)
        {   fprintf(f, "%6u ", n);
            if  (ip_tree[n].mask)
            {   print_ip(f, ip_tree[n].value);
                fprintf(f, "\t       ");
                print_ip(f, ip_tree[n].mask);
                fprintf(f, "\t%6u= %6u<>\n", ip_tree[n].eq, ip_tree[n].neq);
            }else
            {   fprintf(f, " %6uIn, %6uOut\n", ip_tree[n].eq, ip_tree[n].neq);
        }   }
    }else
    {   unsigned nn, m0, m1, m2, m3, i0, i1, i2, i3, eq, neq;
        ip_tree = malloc(sizeof(struct tree_node));
        if  ((f = fdopen(ifile, "rb")) < 0)                 err(EX_NOINPUT, "Can't fdopen input file");
        for (n = 0;; n++)
        {   if  (feof(f)) break;
            fscanf(f, "%u %u.%u.%u.%u %u.%u.%u.%u %u %u\n", &nn, &i0, &i1, &i2, &i3, &m0, &m1, &m2, &m3, &eq, &neq);
            if  (nn != n)
            {   fprintf(stderr, "Input string number wrong %u in string %u\n", nn, n);
                break;
            }
            ip_tree->mask = ((m0 & 0xFF) << 24) |((m1 & 0xFF) << 16) |((m2 & 0xFF) << 8) |(m3 & 0xFF);
            ip_tree->value = ((i0 & 0xFF) << 24) |((i1 & 0xFF) << 16) |((i2 & 0xFF) << 8) |(i3 & 0xFF);
            ip_tree->eq = eq;
            ip_tree->neq = neq;
            write(ofile, ip_tree, sizeof(struct tree_node));
        }
        close(ofile);
    }
    exit(EX_OK);
}
